from typing import Final

# fmt: off
SYMBOLS: Final[dict[str, str]] = {
    "=":  "EQUALS",
    ">":  "GREATER",
    "<":  "LESS",
    "+":  "PLUS",
    "-":  "MINUS",
    "*":  "STAR",
    "/":  "SLASH",
    "!":  "BANG",
    "^":  "XOR",
    ",":  "COMMA",
    ".":  "DOT",
    "(":  "LPAREN",
    ")":  "RPAREN",
    "{":  "LBRACE",
    "}":  "RBRACE",
    "[":  "LBRACKET",
    "]":  "RBRACKET",
    ":":  "COLON",
    "==": "EQUALS_EQUALS",
    "!=": "NOT_EQUALS",
    ">=": "GREATER_EQUALS",
    "<=": "LESS_EQUALS",
    "&&": "AND",
    "||": "OR",
    "++": "PLUS_PLUS",
    "--": "MINUS_MINUS",
    "//": "SLASH_SLASH",
    "**": "STAR_STAR",
    "->": "ARROW",
    ">>": "GREATER_GREATER",
    "<<": "LESS_LESS",
}
# fmt: on

SYMBOL_CHARS: Final[set[str]] = set("".join(SYMBOLS))



from .keyword import KEYWORDS
from .lexer import Lexer
from .symbol import SYMBOL_CHARS, SYMBOLS
from .token import Token, TokenMetadata, TokenType

__all__ = (
    "KEYWORDS",
    "Lexer",
    "SYMBOL_CHARS",
    "SYMBOLS",
    "Token",
    "TokenMetadata",
    "TokenType",
)



from pathlib import Path
from sys import argv

from .frontend.lexer import Lexer


def main() -> None:
    command = argv[1]
    file = argv[2]

    if command == "lex":
        lexer = Lexer(file, Path(file).read_text())

        for token in lexer.tokenise():
            print(token)

    else:
        raise Exception("Unknown command")


if __name__ == "__main__":
    main()



from typing import Optional

from .keyword import KEYWORDS
from .symbol import SYMBOL_CHARS, SYMBOLS
from .token import Token, TokenMetadata, TokenType


class Lexer:
    __slots__ = (
        "_file",
        "_data",
        "_line",
        "_index",
    )

    def __init__(self, file: str, data: str) -> None:
        self._file = file
        self._data = data

        self._line = 1
        self._index = 0

    def _get_token(self, type: TokenType, value: Optional[str] = None) -> Token:
        return Token(type, value, TokenMetadata(self._file, self._line, self._index - (len(value) if value else 0)))

    def _match(self, value: str) -> bool:
        for i, char in enumerate(value):
            if i >= len(self._data) or self._data[self._index + i] != char:
                return False

        return True

    def _peek(self, offset: int = 0, length: int = 1) -> Optional[str]:
        if self._index + offset + length > len(self._data):
            return

        return self._data[self._index + offset : self._index + offset + length]

    def _read_number(self) -> Token:
        number = ""
        decimal = False

        while (char := self._peek()) in set("01234567890."):
            if char == ".":
                if decimal:
                    raise Exception("Invalid number: found unexpected decimal after existing decimal")

                decimal = True

            number += char
            self._index += 1

        return self._get_token(TokenType.NUMBER, number)

    def _read_string(self, quote: str) -> Token:
        string = ""
        escape = False
        closed = False

        while (char := self._peek(1)) is not None:
            if char == "\\":
                escape = True
            elif char == quote and not escape:
                self._index += 1
                closed = True
                break

            string += char
            self._index += 1

        if not closed:
            raise Exception("Unterminated string literal.")

        self._index += 1

        return self._get_token(TokenType.STRING, string)

    def _read_identifier(self) -> Token:
        identifier = ""

        while (char := self._peek()) is not None and char.isalnum():
            identifier += char
            self._index += 1

        if identifier in KEYWORDS:
            return self._get_token(getattr(TokenType, KEYWORDS[identifier]))

        return self._get_token(TokenType.IDENTIFIER, identifier)

    def _read_symbol(self) -> Token:
        symbol = ""

        while (char := self._peek()) is not None and char in SYMBOL_CHARS:
            symbol += char

            if symbol not in SYMBOLS:
                symbol = symbol[:-1]
                break

            self._index += 1

        if symbol in SYMBOLS:
            return self._get_token(getattr(TokenType, SYMBOLS[symbol]))

        raise Exception("Invalid symbol: " + symbol)

    def _read_comment(self) -> Token:
        comment = ""

        while (char := self._peek()) is not None and char != "\n":
            comment += char
            self._index += 1

        self._index += 1

        return self._get_token(TokenType.COMMENT, comment)

    def _read_token(self) -> Token:
        while (char := self._peek()) in [" ", "\n", "\r", "\t", None]:
            if char is None:
                return self._get_token(TokenType.EOF)

            if char == "\n":
                self._line += 1
            self._index += 1

        if char is None:
            return self._get_token(TokenType.EOF)

        if char in SYMBOL_CHARS:
            return self._read_symbol()

        if char.isdigit():
            return self._read_number()

        if char.isalpha():
            return self._read_identifier()

        if char in ["'", '"']:
            return self._read_string(char)

        if char == "#":
            return self._read_comment()

        raise Exception("Invalid character: " + char)

    def tokenise(self) -> list[Token]:
        tokens: list[Token] = []

        while True:
            token = self._read_token()
            tokens.append(token)

            if token.type is TokenType.EOF:
                break

        return tokens



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional


# fmt: off
class TokenType(Enum):
    EOF             = auto()
    IDENTIFIER      = auto()
    NUMBER          = auto()
    STRING          = auto()
    COMMENT         = auto()
    IF              = auto()
    ELSE            = auto()
    WHILE           = auto()
    FOR             = auto()
    RETURN          = auto()
    BREAK           = auto()
    CONTINUE        = auto()
    TRUE            = auto()
    FALSE           = auto()
    NULL            = auto()
    IMPORT          = auto()
    AS              = auto()
    FROM            = auto()
    ASYNC           = auto()
    AWAIT           = auto()
    CLASS           = auto()
    FUNCTION        = auto()
    PUBLIC          = auto()
    EQUALS          = auto()
    GREATER         = auto()
    LESS            = auto()
    PLUS            = auto()
    MINUS           = auto()
    STAR            = auto()
    SLASH           = auto()
    BANG            = auto()
    XOR             = auto()
    COMMA           = auto()
    DOT             = auto()
    LPAREN          = auto()
    RPAREN          = auto()
    LBRACE          = auto()
    RBRACE          = auto()
    LBRACKET        = auto()
    RBRACKET        = auto()
    COLON           = auto()
    EQUALS_EQUALS   = auto()
    NOT_EQUALS      = auto()
    GREATER_EQUALS  = auto()
    LESS_EQUALS     = auto()
    AND             = auto()
    OR              = auto()
    PLUS_PLUS       = auto()
    MINUS_MINUS     = auto()
    SLASH_SLASH     = auto()
    STAR_STAR       = auto()
    ARROW           = auto()
    GREATER_GREATER = auto()
    LESS_LESS       = auto()
# fmt: on


@dataclass(slots=True, frozen=True)
class TokenMetadata:
    file: str
    line: int
    index: int


@dataclass(slots=True, frozen=True)
class Token:
    type: TokenType
    value: Optional[str]
    meta: TokenMetadata



from typing import Final

# fmt: off
KEYWORDS: Final[dict[str, str]] = {
    "if":       "IF",
    "else":     "ELSE",
    "while":    "WHILE",
    "for":      "FOR",
    "return":   "RETURN",
    "break":    "BREAK",
    "continue": "CONTINUE",
    "true":     "TRUE",
    "false":    "FALSE",
    "null":     "NULL",
    "import":   "IMPORT",
    "as":       "AS",
    "from":     "FROM",
    "async":    "ASYNC",
    "await":    "AWAIT",
    "class":    "CLASS",
    "fn":       "FUNCTION",
    "public":   "PUBLIC",
}
# fmt: on



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['volcan',
 'volcan.backend',
 'volcan.frontend',
 'volcan.frontend.lexer',
 'volcan.frontend.parser']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'volcan',
    'version': '0.0.1',
    'description': 'A modern, strongly-typed, general-purpose programming language.',
    'long_description': '# Volcan\n\nA modern, strongly-typed, general-purpose programming language.\n',
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
